#--------------------------------------------------------------------------------------------------------------------------
# Funktions to Read or Genrate the Modbusblock from Wifi-X / TCP 
#
# License:
# Feel free to use this code in your private or business environment or as example for your own code.
# Redistribution is only allowed with no changes.
# For any other kind of usage, please ask me.
# Ralf Amandi, Avarix GmbH
#
# 05.05.2021 Version 1.0
#--------------------------------------------------------------------------------------------------------------------------
import sys
import time

#------------------------------------------------------------------------------
# binReader:
# read/write values from/to the binary Block (bytes or bytearray)
#------------------------------------------------------------------------------
class binReader:
	def __init__(self,data,ofs=0):
		self.data=data
		self.ofs=ofs;
	def getB(self,pos):
		return self.data[pos+self.ofs]
	def getW(self,pos):
		return self.data[pos+self.ofs]*256+self.data[pos+self.ofs+1]
	def getI(self,pos):
		w=self.getW(pos)
		if (w>>15==1):
			w=w-65536
		return w	
	def getI32(self,pos):
		l=self.getW(pos)*65536+self.getW(pos+2)
		if (l>>31==1):
			l=l-4294967296;
		return l
	def getStr(self,pos,cnt):
		s=""
		for i in range(cnt):
			s=s+chr(self.getB(pos+i))
		return s
	def getBytes(self,pos,cnt):
		pos+=self.ofs
		return self.data[pos:pos+cnt]	
	def len(self): # Len from startoffset to end
		return len(self.data)-self.ofs
	def setOfs(self,ofs): # To set the start (Pos=0) to the inner Datablock
		self.ofs=ofs;
	def setB(self,pos,v):
		self.data[pos+self.ofs] = v
	def setW(self,pos,v):
		self.data[pos+self.ofs] = (v >> 8) & 255
		self.data[pos+self.ofs+1] = v & 255
	def Xor(self,pos,cnt,val):
		pos=pos+self.ofs
		l=len(val)
		d=bytearray(self.data)
		v=bytearray(val,"ascii")
		for i in range(cnt):
			d[i+pos] = d[i+pos] ^ v[i % l]
		self.data=d
	def calcCRC(self,pos,cnt): 
		crc = 65535
		pos=pos+self.ofs
		for i in range(cnt):
			crc = crc ^ self.data[i+pos]
			for j in range(8):
				if ((crc & 1) == 1):
					crc = crc >>1
					crc = crc ^0xA001
				else:
					crc = crc >>1
		return crc
		
#------------------------------------------------------------------------------
# class to hold the data auf a ModbusBlock
# The binary block has the structure
# transactId, protocoll,datalen,addr, func, data,crc
#------------------------------------------------------------------------------
class modbusBlock(binReader):
	def __init__(self,dataBlock, xorStr=None):
		binReader.__init__(self,dataBlock)
		self.transactId=self.getW(0)
		self.protocoll=self.getW(2)
		self.datalen=self.getW(4)
		self.addr=self.getB(6)
		self.func=self.getB(7)	# 3=READ HOLDING REGISTERS; 4=READ INPUT REGISTERS
		self.setOfs(8) # Inner block
		#-- check crc --
		crc1 = self.getW(self.datalen-2)
		crc2 = self.calcCRC(-8,self.datalen+6)
		if (crc1 != crc2):
			print("CRC-Error ",hex(crc1), "-",hex(crc2))
		#--
		if (xorStr): self.Xor(0,self.len(),xorStr)

	# returns teh content
	def dump(self):
		h=self.data[0:self.datalen+8].hex()
		lh=len(h)
		sn=""
		if (self.datalen>11):
			sn = self.getStr(0,11)
		txt="-- "+sn+"-"+str(len(self.data))+"--\n"
		txt=txt+" transId: "+str(self.transactId)+", protocoll: "+str(self.protocoll)+", len: "+str(self.datalen)+", addr: "+str(self.addr)+", func: "+ str(self.func)+"\n"
		txt=txt+" DATA: "+h[0:16]+" "+h[16:lh-4]+" "+h[lh-4:lh]+"\n"
		return txt;

	# retrieve the binary block, do an Xor for Growatt
	def getBlock(self, xorStr=None): # Xored with "Growatt" and with CRC
		hd=bytearray([(self.transactId>>8) & 255 ,self.transactId & 255, (self.protocoll>>8) & 255 ,self.protocoll & 255, (self.datalen>>8) & 255 ,self.datalen & 255,  self.addr, self.func])
		hd.extend(self.data[self.ofs:self.ofs+self.datalen-2])
		hd.extend([0,0]) # CRC
		d=binReader(hd,8)
		if (xorStr): d.Xor(0,len(d.data)-8,xorStr)
		crc2 = d.calcCRC(-8,self.datalen+6)
		d.setW(self.datalen-2,crc2)
		return d.data
	
	# generate a responseblock by copying the header and adding the data
	def genResponseBlock(self,data):
		return modbusBlockNew(self.transactId,self.protocoll,self.addr,self.func,bytearray(data))

# create a block from values instat from binary		
class modbusBlockNew(modbusBlock): # generating a new Block
	def __init__(self,transactId, protocoll,addr, func, data1, data2=None):			
		data1= bytearray(data1) # vorsichtshalber kopieren, extend verändert sonst das übergebene array
		if (data2): data1.extend(data2); 
		self.transactId=transactId
		self.protocoll=protocoll
		self.datalen=len(data1)+2;
		self.addr=addr
		self.func=func
		binReader.__init__(self,data1)
		
